//import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import javax.swing.*;
import java.util.*;
import javax.swing.border.*;

public class Client extends JComponent implements Runnable
{	private Socket socket;
	private DataOutputStream dout;
	private DataInputStream din;
	private int playerNumber;
	private boolean gameStarted;
	
	String host;
	String username;
	int port;

	ClientApp capp;
	
	int blockSize=30;
	int blocks=17;
	ImageIcon[] tankUP, tankDOWN, tankLEFT, tankRIGHT;
	ImageIcon block, bullet, coin, loading;
	ImageIcon timefreeze, firepower;
	
	//Flags & Resources for use by PaintComponent
	boolean drawMap=false;
	boolean drawPlayers=false;
	boolean drawPowerUp=false;
	String[] powerUp;
	String[][] map;
	Vector pi;
	String[] hasFired;
	
	public Client( String inputHost, int inputPortNumber, String inputUsername, ClientApp capp)
	{	this.capp=capp;
		host=inputHost;
		username=inputUsername;
		port=inputPortNumber;
		gameStarted=false;
		
		setPreferredSize(new Dimension(510,510));
		loadImages();
		initKeyBindings();
		initPlayerData();
		map=new String[17][17];
		
		try
		{	socket = new Socket( host, port );
			System.out.println( "connected to "+socket );
			din = new DataInputStream( socket.getInputStream() );
			dout = new DataOutputStream( socket.getOutputStream() );
			new Thread( this ).start();
		}
		catch(IOException ie)
		{	ie.printStackTrace();
		}
	}
	public void initPlayerData()
	{	pi=new Vector();
		for(int i=0;i<4;i++)
		{	pi.addElement(new PlayerInfo(blocks+1,blocks+1,"",(i+1)));
		}
	}
	public void loadImages()
	{	tankUP		=new ImageIcon[4];
		tankDOWN	=new ImageIcon[4];
		tankLEFT	=new ImageIcon[4];
		tankRIGHT	=new ImageIcon[4];
		
		for(int i=0;i<4;i++)
		{	tankUP[i]		=new ImageIcon("Tank"+(i+1)+" UP.jpg");
			tankDOWN[i]		=new ImageIcon("Tank"+(i+1)+" DOWN.jpg");
			tankLEFT[i]		=new ImageIcon("Tank"+(i+1)+" LEFT.jpg");
			tankRIGHT[i]	=new ImageIcon("Tank"+(i+1)+" RIGHT.jpg");
		}
		
		block		=new ImageIcon("Block.jpg");
		bullet		=new ImageIcon("Bullet.jpg");
		coin		=new ImageIcon("Coin.jpg");
		loading		=new ImageIcon("Loading.jpg");
		
		timefreeze	=new ImageIcon("Timefreeze.jpg");
		firepower	=new ImageIcon("Firepower.jpg");
	}
	public void initKeyBindings()
	{	Action Up = new AbstractAction()
		{	public void actionPerformed(ActionEvent e)
			{	if(gameStarted)
				{	sendMessage("UP "+playerNumber);
				}
			}
		};
		Action Down = new AbstractAction()
		{	public void actionPerformed(ActionEvent e)
			{	if(gameStarted)
				{	sendMessage("DOWN "+playerNumber);
				}
			}
		};
		Action Left = new AbstractAction()
		{	public void actionPerformed(ActionEvent e)
			{	if(gameStarted)
				{	sendMessage("LEFT "+playerNumber);
				}
			}
		};
		Action Right = new AbstractAction()
		{	public void actionPerformed(ActionEvent e)
			{	if(gameStarted)
				{	sendMessage("RIGHT "+playerNumber);
				}
			}
		};
		Action Fire = new AbstractAction()
		{	public void actionPerformed(ActionEvent e)
			{	if(gameStarted)
				{	sendMessage("FIRE "+playerNumber);
				}
			}
		};
		getInputMap().put(KeyStroke.getKeyStroke("UP"),"Up");
		getInputMap().put(KeyStroke.getKeyStroke("DOWN"),"Down");
		getInputMap().put(KeyStroke.getKeyStroke("LEFT"),"Left");
		getInputMap().put(KeyStroke.getKeyStroke("RIGHT"),"Right");
		getInputMap().put(KeyStroke.getKeyStroke("ENTER"),"Fire");
		
		getActionMap().put("Up",Up);
		getActionMap().put("Down",Down);
		getActionMap().put("Left",Left);
		getActionMap().put("Right",Right);
		getActionMap().put("Fire",Fire);
	}
	public void sendMessage(String message)
	{	try
		{	dout.writeUTF( message );
		}
		catch( IOException ie )
		{	System.out.println( ie );
		}
	}
	
	public void run()
	{	try 
		{	while (true)
			{	String message = din.readUTF();
				System.out.println("recieved message: "+message);
				
				if(message.startsWith("/playerNumber"))	// recieves server-assigned player #
				{	message=message.substring(14);
					System.out.println("PLAYER NUMBER SET TO "+message);
					playerNumber=Integer.parseInt(message);
				}
				else if(message.startsWith("/disableStart")) // recieves the official startgame from server
				{	System.out.println("DISABLING START MENU");
					capp.hide();
				}
				else if(message.startsWith("/startGame")) // recieves the official startgame from server
				{	System.out.println("STARTING GAME");
					gameStarted=true;
				}
				else if(message.startsWith("/mapData"))
				{	message=message.substring(9);
					String[] tempZ=message.split(" ");
					
					String[] temp;						//Map Format: 0->X 1->Y 2->Type
					for(int i=0;i<tempZ.length;i++)
					{	temp=tempZ[i].split(",");
						
						if(temp[2].equals("B"))
							map[Integer.parseInt(temp[0])][Integer.parseInt(temp[1])]="BLOCK";
						else if(temp[2].equals("S"))
							map[Integer.parseInt(temp[0])][Integer.parseInt(temp[1])]="SPACE";
						else if(temp[2].equals("C"))
							map[Integer.parseInt(temp[0])][Integer.parseInt(temp[1])]="COIN";
						else if(temp[2].equals("U"))
							map[Integer.parseInt(temp[0])][Integer.parseInt(temp[1])]="BULLET";
					}	
					
					System.out.println("RECIEVED MAP WITH "+tempZ.length+" ELEMENTS");
					
					drawMap=true;
					repaint();
				}
				else if(message.startsWith("/playerData"))
				{	message=message.substring(12);
					String[] temp=message.split(",");		//0->playerNum 1->X 2->Y 3->Facing
					
					System.out.println("UPDATING INFO OF PLAYER "+temp[0]+"-"+temp[1]+","+temp[2]+" FACING "+temp[3]);
					
					//Updates the previous location of the player to a "SPACE" whenever he moves
					int prevX=((PlayerInfo)pi.elementAt(Integer.parseInt(temp[0])-1)).getX();
					int prevY=((PlayerInfo)pi.elementAt(Integer.parseInt(temp[0])-1)).getY();
					if(prevX<=blocks&&prevY<=blocks)
						map[prevX][prevY]="SPACE";
					
					if(drawPowerUp&&
					   Integer.parseInt(temp[1])==Integer.parseInt(powerUp[0])&&
					   Integer.parseInt(temp[2])==Integer.parseInt(powerUp[1]))
					{	drawPowerUp=false;
					}
					
					((PlayerInfo)pi.elementAt(Integer.parseInt(temp[0])-1)).setX(Integer.parseInt(temp[1]));
					((PlayerInfo)pi.elementAt(Integer.parseInt(temp[0])-1)).setY(Integer.parseInt(temp[2]));
					((PlayerInfo)pi.elementAt(Integer.parseInt(temp[0])-1)).setFacing(temp[3]);
					
					drawPlayers=true;
					repaint();
				}
				else if(message.startsWith("/powerUp"))
				{	message=message.substring(9);
					powerUp=message.split(",");
					
					System.out.println("RECIEVED UPDATE ON MAP "+powerUp[0]+","+powerUp[1]+","+powerUp[2]);
					drawPowerUp=true;
				}
				else if(message.startsWith("/bulletData"))
				{	message=message.substring(12);
				
					if(message.equals("CLEAR"))
					{	System.out.println("CLEARING SCREEN OF BULLETS");
						hasFired=null;
						repaint();
					}
					else
					{	hasFired=message.split(" ");
						System.out.println("RECIEVED UPDATE ON "+hasFired.length+" BULLET(S).");
						repaint();
					}
				}
				else if(message.startsWith("/gameFinished"))
				{	message=message.substring(14);
					
					String[] playerStats=message.split(" ");
					
					String[] scores;
					String scoresText=new String();
					String yourScoreText=new String();
					for(int i=0;i<playerStats.length;i++)	
					{	scores=playerStats[i].split(",");	//0->PlayerNumber 1->Score
						
						if(i!=playerNumber-1)
						{	scoresText=scoresText+"Player "+(i+1)+"'s Score: "+scores[1]+"\n";
						}
						else
						{	yourScoreText="YOUR SCORE: "+scores[1];
						}
					}
					scoresText=scoresText+"\n"+yourScoreText;
					
					JOptionPane.showMessageDialog(this,scoresText,"Game Finished!",JOptionPane.INFORMATION_MESSAGE);
					System.exit(0);
				}
			}
		}
		catch( IOException ie )
		{ 	ie.printStackTrace();
			JOptionPane.showMessageDialog(this,"Server has refused connection. Possibly because it is full, currently hosting a game, or is not running.","Connection Refused",JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		}
	}
	public void paintComponent(Graphics g)
	{	g.setColor(Color.white);
		g.fillRect(0,0,getWidth(),getHeight());
		
		if(!gameStarted)
			g.drawImage(loading.getImage(),0,0,Color.red,null);
		
		if(drawMap)
		{	String[] temp;						//Map Format: 0->X 1->Y 2->Type
			for(int x=0;x<blocks;x++)
			{	for(int y=0;y<blocks;y++)
				{	if(map[x][y].equals("BLOCK"))
						g.drawImage(block.getImage(),x*blockSize,y*blockSize,Color.white,null);
					else if(map[x][y].equals("COIN"))
						g.drawImage(coin.getImage(),x*blockSize,y*blockSize,Color.white,null);
					else if(map[x][y].equals("BULLET"))
						g.drawImage(bullet.getImage(),x*blockSize,y*blockSize,Color.white,null);
				}
			}
		}
		if(drawPlayers)
		{	for(int i=0;i<4;i++)
			{	PlayerInfo temp=(PlayerInfo)pi.elementAt(i);
				
				if(temp.getFacing().equals("UP"))
				{	g.drawImage(tankUP[temp.playerNumber-1].getImage(),temp.getX()*blockSize,temp.getY()*blockSize,Color.white,null);
				}
				else if(temp.getFacing().equals("DOWN"))
				{	g.drawImage(tankDOWN[temp.playerNumber-1].getImage(),temp.getX()*blockSize,temp.getY()*blockSize,Color.white,null);
				}
				else if(temp.getFacing().equals("LEFT"))
				{	g.drawImage(tankLEFT[temp.playerNumber-1].getImage(),temp.getX()*blockSize,temp.getY()*blockSize,Color.white,null);
				}
				else if(temp.getFacing().equals("RIGHT"))
				{	g.drawImage(tankRIGHT[temp.playerNumber-1].getImage(),temp.getX()*blockSize,temp.getY()*blockSize,Color.white,null);
				}
			}
		}
		if(drawPowerUp)
		{	if(powerUp[2].equals("FIREPOWER"))
			{	g.drawImage(firepower.getImage(),Integer.parseInt(powerUp[0])*blockSize,Integer.parseInt(powerUp[1])*blockSize,Color.white,null);
			}
			else if(powerUp[2].equals("TIMEFREEZE"))
			{	g.drawImage(timefreeze.getImage(),Integer.parseInt(powerUp[0])*blockSize,Integer.parseInt(powerUp[1])*blockSize,Color.white,null);
			}
		}
		if(hasFired!=null)
		{	for(int i=0;i<hasFired.length;i++)
			{	String[] temp	=hasFired[i].split(",");
				int x			=Integer.parseInt(temp[0]);
				int y			=Integer.parseInt(temp[1]);
			
				g.drawImage(bullet.getImage(),x*blockSize,y*blockSize,Color.red,null);
			}
		}
	}
}