import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import java.awt.*;
import java.applet.*;
import java.awt.event.*;

public class Server extends JFrame implements ActionListener
{	ServerSocket ss;
	Hashtable playerNumbers=new Hashtable();
	
	JPanel panelNorth,panelEast,panelWest,panelSouth,panelCenter;
	JTextArea messages;
	JTextArea status;
	
	JTextArea input;
	JButton buttonSend;
	
	JScrollPane statusPane;
	int playerCount=0;
	JLabel labelPlayerCount;
	JLabel labelGameStatus;
	
	private java.util.Timer timer;
    private Task countdownStart;

    boolean gameStarted=false;
    
    Vector gameLogic;
    Vector userStreams;
    
    // GAME SETTINGS
    int countdownCtr=10;
    int playerNumberCtr=1;
    int sessionNumberCtr=1;
		
	class Task extends java.util.TimerTask
	{	public void run()
		{	if(countdownCtr>0)
			{	System.out.println("GAME WILL START IN:"+(countdownCtr--));
			}
			else
			{	timer.cancel();
				countdownCtr=10;
				gameStarted=true;
			}
  		}
  	}
	public Server( int port ) throws IOException
	{	initItems();
		setTitle("Server (on Port Number "+port+")");
		gameLogic=new Vector();
		userStreams=new Vector();
		
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setSize(800,600);
		show();
		listen(port);
	}
	private void initItems()
	{	status=new JTextArea();
		status.setEditable(false);
		status.setColumns(20);
		status.setRows(10);
		
		labelPlayerCount=new JLabel("0");
		labelGameStatus=new JLabel("Game Pending");
		labelGameStatus.setForeground(Color.red);
		
		input=new JTextArea();
		buttonSend=new JButton("Send");
		buttonSend.addActionListener(this);
		
		getContentPane().setLayout(new BorderLayout());
		panelNorth = new JPanel();
		panelEast = new JPanel();
		panelWest = new JPanel();
		panelSouth = new JPanel(new BorderLayout());
		panelCenter = new JPanel(new BorderLayout());
		
		statusPane = new JScrollPane(status);
		panelCenter.add("Center",statusPane);
		
		panelNorth.add(new JLabel("Number of Players: "));
		panelNorth.add(labelPlayerCount);
		panelNorth.add(labelGameStatus);
		panelSouth.add("Center",input);
		panelSouth.add("East",buttonSend);
		
		getContentPane().add("North",panelNorth);
		getContentPane().add("Center",panelCenter);
		getContentPane().add("South",panelSouth);
	}
	public void actionPerformed(ActionEvent e)
	{	if(e.getSource()==buttonSend)
		{	sendToAll(input.getText(),sessionNumberCtr);
		}
	}
	static public void main( String args[] ) throws Exception
	{	int port = 3000;
		new Server(port);		
	}
	private void listen( int port ) throws IOException 
	{	ss = new ServerSocket( port );
		ss.setReuseAddress(true);
		
		System.out.println( "Listening on "+ss );
		while (true) 
		{	System.out.println("listening...");
			Socket s = ss.accept();			
			appendStatus("Connection from "+s);

			if(gameStarted||playerNumberCtr==5)
			{	appendStatus("Game has started or is already full. Closing connection.");
				s.close();
			}
			else
			{	addUser(Integer.toString(s.getPort()),s);
				new ServerThread( this, s, sessionNumberCtr );
			}
		}
	}
	void addUser(String port, Socket s)
	{	try
		{	DataOutputStream oout=new DataOutputStream(s.getOutputStream());	
			//useroutputStreams.put(port,oout);
			if(playerNumberCtr==1)	
			{	userStreams.addElement(new UserStreams());
				System.out.println("Creating new set of userStreams");
			}
			((UserStreams)userStreams.elementAt(sessionNumberCtr-1)).addStream(port,s);
			addPlayer();
			System.out.println("Added new user for session "+sessionNumberCtr+" player "+playerNumberCtr);
			sendMessage("/playerNumber "+playerNumberCtr,oout);
			playerNumberCtr++;
		}
		catch(IOException ie)
		{
		}
	}
	public void removeConnection( Socket s, int sessionNumber)
	{	synchronized(userStreams)
		{	appendStatus("Removing connection to ("+s.getPort()+").");
			getUS(sessionNumber).removeStream(s);
			
			/*if(getUS(sessionNumber).isEmpty())
			{	userStreams.removeElementAt(sessionNumber-1);
				gameLogic.elementAt(sessionNumber-1);
			}*/
			removePlayer();

			try
			{	s.close();
			}
			catch( IOException ie )
			{	ie.printStackTrace();
			}
			
			if(userStreams.isEmpty())
			{	appendStatus("Refreshing Game Settings.");
				gameStarted=false;
				gameLogic=new Vector();
				playerNumberCtr=1;
				sessionNumberCtr=1;
				labelGameStatus.setText("Game Pending");
			}
		}
	}
	public void appendStatus(String message)
	{	status.append(message+"\n");
	}
	public void sendMessage(String message, Socket s)
	{	try
		{	for(int i=0;i<userStreams.size();i++)
			{	if(((UserStreams)userStreams.elementAt(i)).getStream(String.valueOf(s.getPort()))!=null)
				{	DataOutputStream oout=((UserStreams)userStreams.elementAt(i)).getStream(String.valueOf(s.getPort()));
					oout.writeUTF(message);
				}
			}
		}
		catch(IOException ie)
		{	ie.printStackTrace();
		}	
	}
	public void sendMessage(String message, DataOutputStream oout)
	{	try
		{	oout.writeUTF(message);
		}
		catch(IOException ie)
		{	ie.printStackTrace();
		}	
	}
	public void sendToAll(String message,int sessionNumber)
	{	try
		{	int ctr=0;
			Enumeration e=((UserStreams)userStreams.elementAt(sessionNumber-1)).getStreams().elements();
			while(e.hasMoreElements())
			{	DataOutputStream oout=(DataOutputStream)e.nextElement();
				oout.writeUTF(message);
				ctr++;
			}
			appendStatus("Message sent to ("+ctr+") users: "+message);
		}
		catch(IOException ie)
		{
		}
	}
	private void addPlayer()
	{	playerCount++;
		labelPlayerCount.setText(Integer.toString(playerCount));
	}
	private void removePlayer()
	{	playerCount--;
		labelPlayerCount.setText(Integer.toString(playerCount));
	}
	public int getPlayerCount()
	{	return playerCount;
	}
	public void startCountdown()
	{	gameStarted=false;
		timer = new java.util.Timer();
		countdownStart = new Task();
		timer.schedule (countdownStart, 0, 800);
		
		// Initialize GameLogic and send MapData & PlayerData during Countdown
		gameLogic.addElement(new GameLogic(this,sessionNumberCtr));
		getGL(sessionNumberCtr).addPlayers(playerNumberCtr-1);
		System.out.println("Created new GameLogic with "+(playerNumberCtr-1)+" players.");
		sendMapData(sessionNumberCtr);
		
		System.out.println("Requesting player datas for session: "+sessionNumberCtr);
		sendPlayerData(sessionNumberCtr);
		System.out.println("Sent playerdata");
		while(!gameStarted){}
		System.out.println("Game Start!");
		labelGameStatus.setText("Game Started");
		sendToAll("/startGame",sessionNumberCtr);
		getGL(sessionNumberCtr).generatePowerUp();
		getGL(sessionNumberCtr).startBulletTask();
		
		
		//Init New Session
		playerNumberCtr=1;
		sessionNumberCtr++;
		gameStarted=false;
	}
	public void sendPlayerData(int sessionNumber)				//format: "playerInfo PLAYERNUM,X,Y,FACING"
	{	for(int i=0;i<playerNumberCtr-1;i++)
		{	PlayerInfo pi=(PlayerInfo)getGL(sessionNumber).getPlayerData(i+1);
			sendToAll("/playerData "+(i+1)+","+pi.getX()+","+pi.getY()+","+pi.getFacing(),sessionNumber);
		}
	}
	public void sendPlayerData(int playerNum,int sessionNumber)	//format: "playerInfo PLAYERNUM,X,Y,FACING"
	{	PlayerInfo pi=(PlayerInfo)getGL(sessionNumber).getPlayerData(playerNum);
		sendToAll("/playerData "+playerNum+","+pi.getX()+","+pi.getY()+","+pi.getFacing(),sessionNumber);
	}
	public void sendMapData(int sessionNumber)
	{	String[][] map=getGL(sessionNumber).getMapData();	
		String message=new String();
		
		for(int x=0;x<17;x++)
		{	for(int y=0;y<17;y++)
			{	message=message+x+","+y+",";
			
				if(map[x][y].equals("BLOCK"))
				{	message=message+"B";
				}
				else if(map[x][y].equals("SPACE"))
				{	message=message+"S";
				}
				else if(map[x][y].equals("COIN"))
				{	message=message+"C";
				}
				else if(map[x][y].startsWith(("PLAYER")))
				{	message=message+"S";
				}
				else if(map[x][y].equals("BULLET"))
				{	message=message+"U";
				}
				
				if(x!=17||y!=17)
				{	message=message+" ";
				}
			}
		}
		sendToAll("/mapData "+message,sessionNumber);
	}
	//Sends a message to ask the client to update a block in its map
	public void updateMapData(int x,int y,String type,int sessionNumber)
	{	sendToAll("/powerUp "+x+","+y+","+type,sessionNumber);
	}
	public GameLogic getGL(int sessionNumber)
	{	return ((GameLogic)gameLogic.elementAt(sessionNumber-1));
	}
	public UserStreams getUS(int sessionNumber)
	{	return ((UserStreams)userStreams.elementAt(sessionNumber-1));
	}
}